"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var jsx_runtime_1 = require("react/jsx-runtime");
var react_1 = require("react");
var MIN_INDEX = 0;
var normalizeValue = function (min, value, max) { return Math.max(Math.min(value, max), min); };
var getPropName = function (axis) { return ({
    minHeight: axis === 'y' ? 'minHeight' : 'minWidth',
    height: axis === 'y' ? 'height' : 'width',
    maxHeight: axis === 'y' ? 'maxHeight' : 'maxWidth',
    top: axis === 'y' ? 'top' : 'left',
    bottom: axis === 'y' ? 'bottom' : 'right',
    scrollHeight: axis === 'y' ? 'scrollHeight' : 'scrollWidth',
    clientHeight: axis === 'y' ? 'clientHeight' : 'clientWidth',
    scrollTop: axis === 'y' ? 'scrollTop' : 'scrollLeft',
    overflowY: axis === 'y' ? 'overflowY' : 'overflowX'
}); };
var getStyleFabric = function (propName) { return function (size) {
    var _a;
    return (_a = {},
        _a[propName.minHeight] = size,
        _a[propName.height] = size,
        _a[propName.maxHeight] = size,
        _a.overflowAnchor = 'none',
        _a.pointerEvents = 'none',
        _a.userSelect = 'none',
        _a.padding = 0,
        _a.margin = 0,
        _a.border = 'none',
        _a);
}; };
var ViewportList = react_1.forwardRef(function (_a, ref) {
    var _b = _a.viewportRef, viewportRef = _b === void 0 ? null : _b, _c = _a.items, items = _c === void 0 ? [] : _c, itemMinSize = _a.itemMinSize, _d = _a.margin, margin = _d === void 0 ? 0 : _d, _e = _a.overscan, overscan = _e === void 0 ? 1 : _e, _f = _a.axis, axis = _f === void 0 ? 'y' : _f, _g = _a.initialIndex, initialIndex = _g === void 0 ? 0 : _g, _h = _a.initialAlignToTop, initialAlignToTop = _h === void 0 ? true : _h, children = _a.children;
    var propName = react_1.useMemo(function () { return getPropName(axis); }, [axis]);
    var getStyle = react_1.useMemo(function () { return getStyleFabric(propName); }, [propName]);
    var maxIndex = items.length - 1;
    var itemMinSizeWithMargin = itemMinSize + margin;
    var overscanSize = overscan * itemMinSizeWithMargin;
    var _j = react_1.useState(function () {
        var normalizedInitialIndex = normalizeValue(MIN_INDEX, initialIndex, maxIndex);
        return [normalizedInitialIndex, normalizedInitialIndex];
    }), _k = _j[0], startIndex = _k[0], endIndex = _k[1], setIndexes = _j[1];
    var topRef = react_1.useRef(null);
    var bottomRef = react_1.useRef(null);
    var variables = react_1.useRef({
        cache: [],
        step: function () { return null; },
        scrollToIndex: startIndex ? { index: startIndex, alignToTop: initialAlignToTop } : null
    });
    var normalizedStartIndex = normalizeValue(MIN_INDEX, startIndex, maxIndex);
    var normalizedEndIndex = normalizeValue(normalizedStartIndex, endIndex, maxIndex);
    var topStyle = react_1.useMemo(function () {
        return getStyle(variables.current.cache
            .slice(MIN_INDEX, normalizedStartIndex)
            .reduce(function (sum, next) { return sum + next; }, normalizedStartIndex * itemMinSizeWithMargin));
    }, [normalizedStartIndex, itemMinSizeWithMargin, getStyle]);
    var bottomStyle = react_1.useMemo(function () {
        return getStyle(variables.current.cache
            .slice(normalizedEndIndex + 1, maxIndex)
            .reduce(function (sum, next) { return sum + next; }, itemMinSizeWithMargin * (maxIndex - normalizedEndIndex)));
    }, [normalizedEndIndex, maxIndex, itemMinSizeWithMargin, getStyle]);
    var slicedItems = react_1.useMemo(function () {
        return items
            .slice(normalizedStartIndex, normalizedEndIndex + 1)
            .map(function (item, index) { return children(item, normalizedStartIndex + index); });
    }, [items, normalizedStartIndex, normalizedEndIndex, children]);
    var scroll = function () {
        if (!(viewportRef === null || viewportRef === void 0 ? void 0 : viewportRef.current)) {
            return;
        }
        var oldScroll = viewportRef.current[propName.scrollTop];
        var oldScrollSize = viewportRef.current[propName.scrollHeight] - viewportRef.current[propName.clientHeight];
        setTimeout(function () {
            var newScrollSize = viewportRef.current[propName.scrollHeight] - viewportRef.current[propName.clientHeight];
            var newScroll = oldScroll + newScrollSize - oldScrollSize;
            if (viewportRef.current[propName.scrollTop] !== newScroll) {
                viewportRef.current[propName.scrollTop] = oldScroll + newScrollSize - oldScrollSize;
            }
        });
    };
    variables.current.step = function () {
        var _a, _b, _c, _d, _e, _f, _g;
        var viewportRect = (_a = viewportRef === null || viewportRef === void 0 ? void 0 : viewportRef.current) === null || _a === void 0 ? void 0 : _a.getBoundingClientRect();
        var topLimit = (viewportRect
            ? normalizeValue(0, viewportRect[propName.top], document.documentElement[propName.clientHeight])
            : 0) - overscanSize;
        var bottomLimit = (viewportRect
            ? normalizeValue(0, viewportRect[propName.bottom], document.documentElement[propName.clientHeight])
            : document.documentElement[propName.clientHeight]) + overscanSize;
        var topElementRect = (_c = (_b = topRef.current) === null || _b === void 0 ? void 0 : _b.nextSibling) === null || _c === void 0 ? void 0 : _c.getBoundingClientRect();
        var bottomElementRect = (_e = (_d = bottomRef.current) === null || _d === void 0 ? void 0 : _d.previousSibling) === null || _e === void 0 ? void 0 : _e.getBoundingClientRect();
        var maxItemsCountInViewPort = Math.ceil((bottomLimit - topLimit) / itemMinSizeWithMargin);
        var nextStartIndex = startIndex;
        var nextEndIndex = endIndex;
        var diff;
        var index;
        var element;
        if (!topElementRect || !bottomElementRect) {
            return;
        }
        if (variables.current.scrollToIndex) {
            var targetIndex = normalizeValue(MIN_INDEX, variables.current.scrollToIndex.index, maxIndex);
            if (targetIndex >= startIndex && targetIndex <= endIndex) {
                index = startIndex;
                element = (_f = topRef.current) === null || _f === void 0 ? void 0 : _f.nextSibling;
                while (!!element && element !== bottomRef.current) {
                    if (index === targetIndex) {
                        element.scrollIntoView(variables.current.scrollToIndex.alignToTop);
                        variables.current.scrollToIndex = null;
                        break;
                    }
                    element = element.nextSibling;
                    ++index;
                }
                return;
            }
            nextStartIndex = targetIndex - maxItemsCountInViewPort;
            nextEndIndex = targetIndex + maxItemsCountInViewPort;
        }
        else if (topElementRect[propName.top] >= bottomLimit) {
            diff = topElementRect[propName.top] - bottomLimit;
            nextEndIndex = startIndex;
            while (diff >= 0 && nextEndIndex > MIN_INDEX) {
                diff -= (variables.current.cache[--nextEndIndex] || 0) + itemMinSizeWithMargin;
            }
            scroll();
            nextStartIndex = nextEndIndex - maxItemsCountInViewPort;
        }
        else if (bottomElementRect[propName.bottom] + margin <= topLimit) {
            diff = topLimit - bottomElementRect[propName.bottom] + margin;
            nextStartIndex = endIndex;
            while (diff >= 0 && nextStartIndex < maxIndex) {
                diff -= (variables.current.cache[++nextStartIndex] || 0) + itemMinSizeWithMargin;
            }
            nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
        }
        else {
            if (topElementRect[propName.bottom] + margin < topLimit) {
                ++nextStartIndex;
            }
            else if (topElementRect[propName.top] >= topLimit) {
                diff = topElementRect[propName.top] - topLimit;
                while (diff >= 0 && nextStartIndex > MIN_INDEX) {
                    diff -= (variables.current.cache[--nextStartIndex] || 0) + itemMinSizeWithMargin;
                }
                scroll();
            }
            if (bottomElementRect[propName.bottom] + margin <= bottomLimit) {
                diff = bottomLimit - bottomElementRect[propName.bottom] - margin;
                while (diff >= 0 && nextEndIndex < maxIndex) {
                    diff -= (variables.current.cache[++nextEndIndex] || 0) + itemMinSizeWithMargin;
                }
            }
            else if (bottomElementRect[propName.top] > bottomLimit) {
                --nextEndIndex;
            }
        }
        nextStartIndex = normalizeValue(MIN_INDEX, nextStartIndex, maxIndex);
        nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);
        if (nextStartIndex !== startIndex || nextEndIndex !== endIndex) {
            index = startIndex;
            element = (_g = topRef.current) === null || _g === void 0 ? void 0 : _g.nextSibling;
            while (!!element && index <= endIndex && element !== bottomRef.current) {
                variables.current.cache[index++] = element[propName.clientHeight] - itemMinSize;
                element = element.nextSibling;
            }
            setIndexes([nextStartIndex, nextEndIndex]);
        }
    };
    react_1.useImperativeHandle(ref, function () { return ({
        scrollToIndex: function (index, alignToTop) {
            if (index === void 0) { index = -1; }
            if (alignToTop === void 0) { alignToTop = true; }
            variables.current.scrollToIndex = { index: index, alignToTop: alignToTop };
        }
    }); }, []);
    react_1.useEffect(function () {
        var frameId;
        var frame = function () {
            frameId = requestAnimationFrame(frame);
            variables.current.step();
        };
        frame();
        return function () {
            cancelAnimationFrame(frameId);
        };
    }, []);
    return (jsx_runtime_1.jsxs(react_1.Fragment, { children: [jsx_runtime_1.jsx("div", { ref: topRef, style: topStyle }, void 0), slicedItems, jsx_runtime_1.jsx("div", { ref: bottomRef, style: bottomStyle }, void 0)] }, void 0));
});
exports.default = ViewportList;
//# sourceMappingURL=index.js.map